import os
from functools import cache
from typing import Annotated, Any, Self, cast
import logging

from pydantic import BaseModel, Field, model_validator
from pydantic_yaml import parse_yaml_file_as
from fastapi import Depends

from xmpp_api.exceptions.base import ConfigurationException


log = logging.getLogger("config")


class _ComponentConfig(BaseModel):
    # The JID of the component
    jid: str

    # Address of server's component port
    server: str

    # The component's secret.
    secret_plain: str | None = Field(default=None, alias="secret")

    # Optional file to read the secret from
    secret_file: str | None = Field(default=None)

    @model_validator(mode="after")
    def validate_secret(self) -> Self:
        if self.secret_plain is None and self.secret_file is None:
            raise ConfigurationException("No component secret is specified")
        if self.secret_plain is not None and self.secret_file is not None:
            log.warn(
                "Both component.secret and component.secret_file specified! component.secret_file takes precedence"
            )
        return self

    @property
    def secret(self) -> str:
        """
        The component secret.

        The backing may be None if secret_file is used instead. secret_plain is, however, replaced
        with that file's contents during configuration loading.
        """
        return cast(str, self.secret_plain)


class _DatabaseConfig(BaseModel):
    # The URI to connect to
    uri_plain: str | None = Field(default=None, alias="uri")

    # The file to read the database URI from
    uri_file: str | None = Field(default=None)

    # sqlalchemy options for the database connection
    connect_args: dict[str, Any] = Field(default_factory=dict)

    @model_validator(mode="after")
    def validate_secret(self) -> Self:
        if self.uri_plain is None and self.uri_file is None:
            raise ConfigurationException("No database URI is specified")
        if self.uri_plain is not None and self.uri_file is not None:
            log.warn(
                "Both database.uri and database.uri_file specified! database.uri_file takes precedence"
            )
        return self

    @property
    def uri(self) -> str:
        """
        The database URI.

        The backing may be None if uri_file is used instead. uri_plain is, however, replaced
        with that file's contents during configuration loading.
        """

        return cast(str, self.uri_plain)


class _SentryConfig(BaseModel):
    # The Sentry DSN.
    dsn: str


class _Config(BaseModel):
    # Database config
    database: _DatabaseConfig

    # Component configuration
    component: _ComponentConfig

    # Optional Sentry config
    sentry: _SentryConfig | None


@cache
def load_config() -> _Config:
    """
    Load the application config
    """

    config_path = "/etc/xmpp-api/config.yaml"
    if (env_config_file := os.environ.get("XMPP_API__CONFIG_FILE")) is not None:
        config_path = env_config_file

    log.info(f"Reading configuration from {config_path}")
    with open(config_path, "r", encoding="utf8") as f:
        config = parse_yaml_file_as(_Config, f)

    # Read the secret from a file, if specified
    if config.component.secret_file is not None:
        log.info("Reading component secret from %s", config.component.secret_file)
        with open(config.component.secret_file, "r", encoding="utf8") as f:
            config.component.secret_plain = f.read().strip().replace("\n", "")

    # Read the database URI from a file, if specified
    if config.database.uri_file is not None:
        log.info("Reading database URI from %s", config.database.uri_file)
        with open(config.database.uri_file, "r", encoding="utf8") as f:
            config.database.uri_plain = f.read().strip().replace("\n", "")

    return config


ConfigDep = Annotated[_Config, Depends(load_config)]
